@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2014-2016 gnome-mpv
+ * Copyright (c) 2014-2017 gnome-mpv
  *
  * This file is part of GNOME MPV.
  *
@@ -24,6 +24,13 @@
 #include "gmpv_control_box.h"
 #include "gmpv_seek_bar.h"
 
+enum
+{
+	PROP_0,
+	PROP_VOLUME,
+	N_PROPERTIES
+};
+
 struct _GmpvControlBox
 {
 	GtkBox parent_instance;
@@ -36,6 +43,7 @@ struct _GmpvControlBox
 	GtkWidget *volume_button;
 	GtkWidget *fullscreen_button;
 	GtkWidget *seek_bar;
+	gdouble volume;
 };
 
 struct _GmpvControlBoxClass
@@ -43,6 +51,22 @@ struct _GmpvControlBoxClass
 	GtkBoxClass parent_class;
 };
 
+static void set_property(	GObject *object,
+				guint property_id,
+				const GValue *value,
+				GParamSpec *pspec );
+static void get_property(	GObject *object,
+				guint property_id,
+				GValue *value,
+				GParamSpec *pspec );
+static gboolean gtk_to_mpv_volume(	GBinding *binding,
+					const GValue *from_value,
+					GValue *to_value,
+					gpointer data );
+static gboolean mpv_to_gtk_volume(	GBinding *binding,
+					const GValue *from_value,
+					GValue *to_value,
+					gpointer data );
 static void seek_handler(	GmpvSeekBar *seek_bar,
 				gdouble value,
 				gpointer data );
@@ -50,9 +74,68 @@ static void volume_changed_handler(	GtkVolumeButton *button,
 					gdouble value,
 					gpointer data );
 static void simple_signal_handler(GtkWidget *widget, gpointer data);
+static void init_button(	GtkWidget *button,
+				const gchar *icon_name,
+				const gchar *tooltip_text );
 
 G_DEFINE_TYPE(GmpvControlBox, gmpv_control_box, GTK_TYPE_BOX)
 
+static void set_property(	GObject *object,
+				guint property_id,
+				const GValue *value,
+				GParamSpec *pspec )
+{
+	GmpvControlBox *self = GMPV_CONTROL_BOX(object);
+
+	if(property_id == PROP_VOLUME)
+	{
+		self->volume = g_value_get_double(value);
+	}
+	else
+	{
+		G_OBJECT_WARN_INVALID_PROPERTY_ID(object, property_id, pspec);
+	}
+}
+
+static void get_property(	GObject *object,
+				guint property_id,
+				GValue *value,
+				GParamSpec *pspec )
+{
+	GmpvControlBox *self = GMPV_CONTROL_BOX(object);
+
+	if(property_id == PROP_VOLUME)
+	{
+		g_value_set_double(value, self->volume);
+	}
+	else
+	{
+		G_OBJECT_WARN_INVALID_PROPERTY_ID(object, property_id, pspec);
+	}
+}
+
+static gboolean gtk_to_mpv_volume(	GBinding *binding,
+					const GValue *from_value,
+					GValue *to_value,
+					gpointer data )
+{
+	gdouble from = g_value_get_double(from_value);
+	g_value_set_double(to_value, from*100);
+
+	return TRUE;
+}
+
+static gboolean mpv_to_gtk_volume(	GBinding *binding,
+					const GValue *from_value,
+					GValue *to_value,
+					gpointer data )
+{
+	gdouble from = g_value_get_double(from_value);
+	g_value_set_double(to_value, from/100.0);
+
+	return TRUE;
+}
+
 static void seek_handler(	GmpvSeekBar *seek_bar,
 				gdouble value,
 				gpointer data )
@@ -64,7 +147,7 @@ static void volume_changed_handler(	GtkVolumeButton *button,
 					gdouble value,
 					gpointer data )
 {
-	g_signal_emit_by_name(data, "volume-changed", value);
+	g_signal_emit_by_name(data, "volume-changed", value*100.0);
 }
 
 static void simple_signal_handler(GtkWidget *widget, gpointer data)
@@ -95,6 +178,19 @@ static void simple_signal_handler(GtkWidget *widget, gpointer data)
 	}
 }
 
+static void init_button(	GtkWidget *button,
+				const gchar *icon_name,
+				const gchar *tooltip_text )
+{
+	GtkWidget *icon =	gtk_image_new_from_icon_name
+				(icon_name, GTK_ICON_SIZE_BUTTON);
+
+	gtk_widget_set_tooltip_text(button, tooltip_text);
+	g_object_set(button, "relief", GTK_RELIEF_NONE, NULL);
+	gtk_widget_set_can_focus(button, FALSE);
+	gtk_button_set_image(GTK_BUTTON(button), icon);
+}
+
 static void gmpv_control_box_class_init(GmpvControlBoxClass *klass)
 {
 	/* Names of signals that have no parameter and return nothing */
@@ -106,6 +202,21 @@ static void gmpv_control_box_class_init(GmpvControlBoxClass *klass)
 						"next-button-clicked",
 						"fullscreen-button-clicked",
 						NULL };
+	GObjectClass *object_class = G_OBJECT_CLASS(klass);
+	GParamSpec *pspec = NULL;
+
+	object_class->set_property = set_property;
+	object_class->get_property = get_property;
+
+	pspec = g_param_spec_double
+		(	"volume",
+			"Volume",
+			"The value of the volume button",
+			0.0,
+			G_MAXDOUBLE,
+			0.0,
+			G_PARAM_READWRITE );
+	g_object_class_install_property(object_class, PROP_VOLUME, pspec);
 
 	for(gint i = 0; simple_signals[i]; i++)
 	{
@@ -145,109 +256,62 @@ static void gmpv_control_box_class_init(GmpvControlBoxClass *klass)
 
 static void gmpv_control_box_init(GmpvControlBox *box)
 {
-	GtkWidget *play_icon;
-	GtkWidget *stop_icon;
-	GtkWidget *forward_icon;
-	GtkWidget *rewind_icon;
-	GtkWidget *previous_icon;
-	GtkWidget *next_icon;
-	GtkWidget *fullscreen_icon;
-
-	box->play_button = gtk_button_new_with_label(NULL);
-	box->stop_button = gtk_button_new_with_label(NULL);
-	box->forward_button = gtk_button_new_with_label(NULL);
-	box->rewind_button = gtk_button_new_with_label(NULL);
-	box->next_button = gtk_button_new_with_label(NULL);
-	box->previous_button = gtk_button_new_with_label(NULL);
-	box->fullscreen_button = gtk_button_new_with_label(NULL);
+	box->play_button = gtk_button_new();
+	box->stop_button = gtk_button_new();
+	box->forward_button = gtk_button_new();
+	box->rewind_button = gtk_button_new();
+	box->next_button = gtk_button_new();
+	box->previous_button = gtk_button_new();
+	box->fullscreen_button = gtk_button_new();
 	box->volume_button = gtk_volume_button_new();
 	box->seek_bar = gmpv_seek_bar_new();
 
-	play_icon
-		= gtk_image_new_from_icon_name
-			("media-playback-start-symbolic", GTK_ICON_SIZE_BUTTON);
-	stop_icon
-		= gtk_image_new_from_icon_name
-			("media-playback-stop-symbolic", GTK_ICON_SIZE_BUTTON);
-	forward_icon
-		= gtk_image_new_from_icon_name
-			("media-seek-forward-symbolic", GTK_ICON_SIZE_BUTTON);
-	rewind_icon
-		= gtk_image_new_from_icon_name
-			("media-seek-backward-symbolic", GTK_ICON_SIZE_BUTTON);
-	next_icon
-		= gtk_image_new_from_icon_name
-			("media-skip-forward-symbolic", GTK_ICON_SIZE_BUTTON);
-	previous_icon
-		= gtk_image_new_from_icon_name
-			("media-skip-backward-symbolic", GTK_ICON_SIZE_BUTTON);
-	fullscreen_icon
-		= gtk_image_new_from_icon_name
-			("view-fullscreen-symbolic", GTK_ICON_SIZE_BUTTON);
+	init_button(	box->play_button,
+			"media-playback-start-symbolic",
+			_("Play") );
+	init_button(	box->stop_button,
+			"media-playback-stop-symbolic",
+			_("Stop") );
+	init_button(	box->forward_button,
+			"media-seek-forward-symbolic",
+			_("Forward") );
+	init_button(	box->rewind_button,
+			"media-seek-backward-symbolic",
+			_("Rewind") );
+	init_button(	box->next_button,
+			"media-skip-forward-symbolic",
+			_("Next Chapter") );
+	init_button(	box->previous_button,
+			"media-skip-backward-symbolic",
+			_("Previous Chapter") );
+	init_button(	box->fullscreen_button,
+			"view-fullscreen-symbolic",
+			_("Toggle Fullscreen") );
 
 	gtk_style_context_add_class
 		(	gtk_widget_get_style_context(GTK_WIDGET(box)),
 			GTK_STYLE_CLASS_BACKGROUND );
 
-	gtk_widget_set_tooltip_text(box->play_button, _("Play"));
-	gtk_widget_set_tooltip_text(box->stop_button, _("Stop"));
-	gtk_widget_set_tooltip_text(box->forward_button, _("Forward"));
-	gtk_widget_set_tooltip_text(box->rewind_button, _("Rewind"));
-	gtk_widget_set_tooltip_text(box->next_button, _("Next Chapter"));
-	gtk_widget_set_tooltip_text(box->previous_button, _("Previous Chapter"));
-	gtk_widget_set_tooltip_text(box->fullscreen_button, _("Toggle Fullscreen"));
-
-	g_object_set(box->play_button, "relief", GTK_RELIEF_NONE, NULL);
-	g_object_set(box->stop_button, "relief", GTK_RELIEF_NONE, NULL);
-	g_object_set(box->forward_button, "relief", GTK_RELIEF_NONE, NULL);
-	g_object_set(box->rewind_button, "relief", GTK_RELIEF_NONE, NULL);
-	g_object_set(box->next_button, "relief", GTK_RELIEF_NONE, NULL);
-	g_object_set(box->previous_button, "relief", GTK_RELIEF_NONE, NULL);
-	g_object_set(box->fullscreen_button, "relief", GTK_RELIEF_NONE, NULL);
-
-	gtk_widget_set_can_focus(box->previous_button, FALSE);
-	gtk_widget_set_can_focus(box->rewind_button, FALSE);
-	gtk_widget_set_can_focus(box->play_button, FALSE);
-	gtk_widget_set_can_focus(box->stop_button, FALSE);
-	gtk_widget_set_can_focus(box->forward_button, FALSE);
-	gtk_widget_set_can_focus(box->next_button, FALSE);
-	gtk_widget_set_can_focus(box->seek_bar, FALSE);
 	gtk_widget_set_can_focus(box->volume_button, FALSE);
-	gtk_widget_set_can_focus(box->fullscreen_button, FALSE);
+	gtk_widget_set_can_focus(box->seek_bar, FALSE);
 
-	gtk_button_set_image
-		(GTK_BUTTON(box->play_button), play_icon);
-	gtk_button_set_image
-		(GTK_BUTTON(box->stop_button), stop_icon);
-	gtk_button_set_image
-		(GTK_BUTTON(box->forward_button), forward_icon);
-	gtk_button_set_image
-		(GTK_BUTTON(box->rewind_button), rewind_icon);
-	gtk_button_set_image
-		(GTK_BUTTON(box->next_button), next_icon);
-	gtk_button_set_image
-		(GTK_BUTTON(box->previous_button), previous_icon);
-	gtk_button_set_image
-		(GTK_BUTTON(box->fullscreen_button), fullscreen_icon);
+	gtk_container_add(GTK_CONTAINER(box), box->previous_button);
+	gtk_container_add(GTK_CONTAINER(box), box->rewind_button);
+	gtk_container_add(GTK_CONTAINER(box), box->play_button);
+	gtk_container_add(GTK_CONTAINER(box), box->stop_button);
+	gtk_container_add(GTK_CONTAINER(box), box->forward_button);
+	gtk_container_add(GTK_CONTAINER(box), box->next_button);
+	gtk_box_pack_start(GTK_BOX(box), box->seek_bar, TRUE, TRUE, 0);
+	gtk_container_add(GTK_CONTAINER(box), box->volume_button);
+	gtk_container_add(GTK_CONTAINER(box), box->fullscreen_button);
 
-	gtk_container_add
-		(GTK_CONTAINER(box), box->previous_button);
-	gtk_container_add
-		(GTK_CONTAINER(box), box->rewind_button);
-	gtk_container_add
-		(GTK_CONTAINER(box), box->play_button);
-	gtk_container_add
-		(GTK_CONTAINER(box), box->stop_button);
-	gtk_container_add
-		(GTK_CONTAINER(box), box->forward_button);
-	gtk_container_add
-		(GTK_CONTAINER(box), box->next_button);
-	gtk_box_pack_start
-		(GTK_BOX(box), box->seek_bar, TRUE, TRUE, 0);
-	gtk_container_add
-		(GTK_CONTAINER(box), box->volume_button);
-	gtk_container_add
-		(GTK_CONTAINER(box), box->fullscreen_button);
+	g_object_bind_property_full(	box->volume_button, "value",
+					box, "volume",
+					G_BINDING_BIDIRECTIONAL,
+					gtk_to_mpv_volume,
+					mpv_to_gtk_volume,
+					NULL,
+					NULL );
 
 	g_signal_connect(	box,
 				"button-press-event",
@@ -335,7 +399,7 @@ void gmpv_control_box_set_volume(GmpvControlBox *box, gdouble volume)
 	g_signal_handlers_block_by_func(box, volume_changed_handler, box);
 
 	gtk_scale_button_set_value
-		(GTK_SCALE_BUTTON(box->volume_button), volume);
+		(GTK_SCALE_BUTTON(box->volume_button), volume/100.0);
 
 	g_signal_handlers_unblock_by_func(box, volume_changed_handler, box);
 }
@@ -354,36 +418,31 @@ gboolean gmpv_control_box_get_volume_popup_visible(GmpvControlBox *box)
 
 void gmpv_control_box_set_playing_state(GmpvControlBox *box, gboolean playing)
 {
-	GtkWidget *play_icon;
-	const gchar *tooltip;
+	GtkWidget *image = gtk_button_get_image(GTK_BUTTON(box->play_button));
+	const gchar *tooltip = playing?_("Pause"):_("Play");
 
-	play_icon = gtk_image_new_from_icon_name
-			( 	playing
-				?"media-playback-pause-symbolic"
-				:"media-playback-start-symbolic",
-				GTK_ICON_SIZE_BUTTON );
-	tooltip = playing?_("Pause"):_("Play");
+	gtk_image_set_from_icon_name( 	GTK_IMAGE(image),
+					playing?
+					"media-playback-pause-symbolic":
+					"media-playback-start-symbolic",
+					GTK_ICON_SIZE_BUTTON );
 
-	gtk_button_set_image(GTK_BUTTON(box->play_button), play_icon);
 	gtk_widget_set_tooltip_text(box->play_button, tooltip);
 }
 
 void gmpv_control_box_set_fullscreen_state(	GmpvControlBox *box,
 						gboolean fullscreen )
 {
-	GtkWidget *fullscreen_icon;
+	GtkWidget *image = gtk_button_get_image(GTK_BUTTON(box->fullscreen_button));
 
-	fullscreen_icon = gtk_image_new_from_icon_name
-				(	fullscreen
-					?"view-restore-symbolic"
-					:"view-fullscreen-symbolic",
+	gtk_image_set_from_icon_name(	GTK_IMAGE(image),
+					fullscreen?
+					"view-restore-symbolic":
+					"view-fullscreen-symbolic",
 					GTK_ICON_SIZE_BUTTON );
-
-	gtk_button_set_image(	GTK_BUTTON(box->fullscreen_button),
-				fullscreen_icon );
 }
 
-void gmpv_control_box_set_fullscreen_btn_visible(	GmpvControlBox *box,
+void gmpv_control_box_set_fullscreen_button_visible(	GmpvControlBox *box,
 							gboolean value )
 {
 	gtk_widget_set_visible(box->fullscreen_button, value);
