@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2014-2016 gnome-mpv
+ * Copyright (c) 2014-2017 gnome-mpv
  *
  * This file is part of GNOME MPV.
  *
@@ -27,7 +27,6 @@
 #include <gdk/gdk.h>
 #include <stdlib.h>
 #include <string.h>
-#include <execinfo.h>
 
 #include <epoxy/gl.h>
 #ifdef GDK_WINDOWING_X11
@@ -49,7 +48,7 @@
 #include "gmpv_mpv_opt.h"
 #include "gmpv_common.h"
 #include "gmpv_def.h"
-#include "gmpv_playlist.h"
+#include "gmpv_marshal.h"
 
 static void *GLAPIENTRY glMPGetNativeDisplay(const gchar *name);
 static void *get_proc_address(void *fn_ctx, const gchar *name);
@@ -61,13 +60,19 @@ static void get_inst_property(	GObject *object,
 				guint property_id,
 				GValue *value,
 				GParamSpec *pspec );
+static void load_from_playlist(GmpvMpv *mpv);
+static void load_scripts(GmpvMpv *mpv);
 static void wakeup_callback(void *data);
+static GmpvPlaylistEntry *parse_playlist_entry(mpv_node_list *node);
+static GmpvTrack *parse_track_entry(mpv_node_list *node);
 static void mpv_prop_change_handler(GmpvMpv *mpv, mpv_event_property* prop);
 static gboolean mpv_event_handler(gpointer data);
-static void update_playlist(GmpvMpv *mpv);
 static gint apply_args(mpv_handle *mpv_ctx, gchar *args);
 static void log_handler(GmpvMpv *mpv, mpv_event_log_message* message);
-static void load_input_conf(GmpvMpv *mpv, const gchar *input_conf);
+static void add_file_to_playlist(GmpvMpv *mpv, const gchar *uri);
+static void update_playlist(GmpvMpv *mpv);
+static void update_metadata(GmpvMpv *mpv);
+static void update_track_list(GmpvMpv *mpv);
 
 G_DEFINE_TYPE(GmpvMpv, gmpv_mpv, G_TYPE_OBJECT)
 
@@ -121,10 +126,6 @@ static void set_inst_property(	GObject *object,
 	{
 		self->wid = g_value_get_int64(value);
 	}
-	else if(property_id == PROP_PLAYLIST)
-	{
-		self->playlist = g_value_get_pointer(value);
-	}
 	else
 	{
 		G_OBJECT_WARN_INVALID_PROPERTY_ID(object, property_id, pspec);
@@ -142,21 +143,131 @@ static void get_inst_property(	GObject *object,
 	{
 		g_value_set_int64(value, self->wid);
 	}
-	else if(property_id == PROP_PLAYLIST)
-	{
-		g_value_set_pointer(value, self->playlist);
-	}
 	else
 	{
 		G_OBJECT_WARN_INVALID_PROPERTY_ID(object, property_id, pspec);
 	}
 }
 
+static void load_from_playlist(GmpvMpv *mpv)
+{
+	GPtrArray * playlist = mpv->playlist;
+
+	for(guint i = 0; playlist && i < playlist->len; i++)
+	{
+		GmpvPlaylistEntry *entry;
+
+		entry = g_ptr_array_index(playlist, i);
+
+		/* Do not append on first iteration */
+		gmpv_mpv_load_file(mpv, entry->filename, i != 0);
+	}
+}
+
+static void load_scripts(GmpvMpv *mpv)
+{
+	gchar *path = get_scripts_dir_path();
+	GDir *dir = g_dir_open(path, 0, NULL);
+
+	if(dir)
+	{
+		const gchar *name;
+
+		do
+		{
+			gchar *full_path;
+
+			name = g_dir_read_name(dir);
+			full_path = g_build_filename(path, name, NULL);
+
+			if(g_file_test(full_path, G_FILE_TEST_IS_REGULAR))
+			{
+				const gchar *cmd[]
+					= {"load-script", full_path, NULL};
+
+				g_info("Loading script: %s", full_path);
+				mpv_command(mpv->mpv_ctx, cmd);
+			}
+
+			g_free(full_path);
+		}
+		while(name);
+
+		g_dir_close(dir);
+	}
+	else
+	{
+		g_warning("Failed to open scripts directory: %s", path);
+	}
+
+	g_free(path);
+}
+
 static void wakeup_callback(void *data)
 {
 	g_idle_add((GSourceFunc)mpv_event_handler, data);
 }
 
+static GmpvPlaylistEntry *parse_playlist_entry(mpv_node_list *node)
+{
+	const gchar *filename = NULL;
+	const gchar *title = NULL;
+
+	for(gint i = 0; i < node->num; i++)
+	{
+		if(g_strcmp0(node->keys[i], "filename") == 0)
+		{
+			filename = node->values[i].u.string;
+		}
+		else if(g_strcmp0(node->keys[i], "title") == 0)
+		{
+			title = node->values[i].u.string;
+		}
+	}
+
+	return gmpv_playlist_entry_new(filename, title);
+}
+
+static GmpvTrack *parse_track_entry(mpv_node_list *node)
+{
+	GmpvTrack *entry = gmpv_track_new();
+
+	for(gint i = 0; i < node->num; i++)
+	{
+		if(g_strcmp0(node->keys[i], "type") == 0)
+		{
+			const gchar *type = node->values[i].u.string;
+
+			if(g_strcmp0(type, "audio") == 0)
+			{
+				entry->type = TRACK_TYPE_AUDIO;
+			}
+			else if(g_strcmp0(type, "video") == 0)
+			{
+				entry->type = TRACK_TYPE_VIDEO;
+			}
+			else if(g_strcmp0(type, "sub") == 0)
+			{
+				entry->type = TRACK_TYPE_SUBTITLE;
+			}
+		}
+		else if(g_strcmp0(node->keys[i], "title") == 0)
+		{
+			entry->title = g_strdup(node->values[i].u.string);
+		}
+		else if(g_strcmp0(node->keys[i], "lang") == 0)
+		{
+			entry->lang = g_strdup(node->values[i].u.string);
+		}
+		else if(g_strcmp0(node->keys[i], "id") == 0)
+		{
+			entry->id = node->values[i].u.int64;
+		}
+	}
+
+	return entry;
+}
+
 static void mpv_prop_change_handler(GmpvMpv *mpv, mpv_event_property* prop)
 {
 	g_debug("Received mpv property change event for \"%s\"", prop->name);
@@ -172,9 +283,56 @@ static void mpv_prop_change_handler(GmpvMpv *mpv, mpv_event_property* prop)
 					MPV_FORMAT_FLAG,
 					&idle_active );
 
-		if(idle_active && !mpv->state.paused)
+		if(idle_active && !mpv->state.paused && !mpv->init_vo_config)
 		{
-			gmpv_mpv_load(mpv, NULL, FALSE, TRUE);
+			load_from_playlist(mpv);
+		}
+	}
+	else if(g_strcmp0(prop->name, "playlist") == 0)
+	{
+		gint64 playlist_count  = 0;
+		gboolean idle_active = FALSE;
+		gboolean was_empty = FALSE;
+
+		mpv_get_property(	mpv->mpv_ctx,
+					"playlist-count",
+					MPV_FORMAT_INT64,
+					&playlist_count );
+		mpv_get_property(	mpv->mpv_ctx,
+					"idle-active",
+					MPV_FORMAT_FLAG,
+					&idle_active );
+
+		was_empty = (mpv->playlist->len == 0);
+
+		if(!idle_active)
+		{
+			update_playlist(mpv);
+		}
+
+		/* Check if we're transitioning from empty playlist to non-empty
+		 * playlist.
+		 */
+		if(was_empty && mpv->playlist->len > 0)
+		{
+			gmpv_mpv_set_property_flag(mpv, "pause", FALSE);
+		}
+	}
+	else if(g_strcmp0(prop->name, "metadata") == 0)
+	{
+		update_metadata(mpv);
+	}
+	else if(g_strcmp0(prop->name, "track-list") == 0)
+	{
+		update_track_list(mpv);
+	}
+	else if(g_strcmp0(prop->name, "vo-configured") == 0)
+	{
+		if(mpv->init_vo_config)
+		{
+			mpv->init_vo_config = FALSE;
+			load_from_playlist(mpv);
+			load_scripts(mpv);
 		}
 	}
 }
@@ -202,7 +360,8 @@ static gboolean mpv_event_handler(gpointer data)
 
 			g_signal_emit_by_name(	mpv,
 						"mpv-prop-change",
-						prop->name );
+						prop->name,
+						prop->data );
 		}
 		else if(event->event_id == MPV_EVENT_IDLE)
 		{
@@ -212,7 +371,11 @@ static gboolean mpv_event_handler(gpointer data)
 
 				gmpv_mpv_set_property_flag
 					(mpv, "pause", TRUE);
-				gmpv_playlist_reset(mpv->playlist);
+			}
+
+			if(!mpv->state.init_load && !mpv->state.loaded)
+			{
+				g_signal_emit_by_name(mpv, "idle");
 			}
 
 			mpv->state.init_load = FALSE;
@@ -222,7 +385,7 @@ static gboolean mpv_event_handler(gpointer data)
 			mpv->state.loaded = TRUE;
 			mpv->state.init_load = FALSE;
 
-			update_playlist(mpv);
+			g_signal_emit_by_name(mpv, "load");
 		}
 		else if(event->event_id == MPV_EVENT_END_FILE)
 		{
@@ -246,9 +409,8 @@ static gboolean mpv_event_handler(gpointer data)
 						"abnormally. Reason: %s."),
 						err );
 
-				gmpv_mpv_set_property_flag
-					(mpv, "pause", TRUE);
-				g_signal_emit_by_name(mpv, "mpv-error", msg);
+				gmpv_mpv_set_property_flag(mpv, "pause", TRUE);
+				g_signal_emit_by_name(mpv, "error", msg);
 
 				g_free(msg);
 			}
@@ -258,6 +420,28 @@ static gboolean mpv_event_handler(gpointer data)
 			if(mpv->state.new_file)
 			{
 				gmpv_mpv_opt_handle_autofit(mpv);
+
+				g_signal_emit_by_name(	mpv,
+							"autofit",
+							mpv->autofit_ratio );
+			}
+		}
+		else if(event->event_id == MPV_EVENT_START_FILE)
+		{
+			gboolean vo_configured = FALSE;
+
+			mpv_get_property(	mpv->mpv_ctx,
+						"vo-configured",
+						MPV_FORMAT_FLAG,
+						&vo_configured );
+
+			/* If the vo is not configured yet, save the content of
+			 * mpv's playlist in mpv->playlist. This will be loaded
+			 * again when the vo is configured.
+			 */
+			if(!vo_configured)
+			{
+				update_playlist(mpv);
 			}
 		}
 		else if(event->event_id == MPV_EVENT_PLAYBACK_RESTART)
@@ -268,147 +452,36 @@ static gboolean mpv_event_handler(gpointer data)
 		{
 			log_handler(mpv, event->data);
 		}
-		else if(event->event_id == MPV_EVENT_SHUTDOWN
-		|| event->event_id == MPV_EVENT_NONE)
+		else if(event->event_id == MPV_EVENT_CLIENT_MESSAGE)
+		{
+			mpv_event_client_message *event_cmsg = event->data;
+			gchar* msg = strnjoinv(	" ",
+						event_cmsg->args,
+						(gsize)event_cmsg->num_args );
+
+			g_signal_emit_by_name(mpv, "message", msg);
+			g_free(msg);
+		}
+		else if(event->event_id == MPV_EVENT_SHUTDOWN)
+		{
+			g_signal_emit_by_name(mpv, "shutdown");
+
+			done = TRUE;
+		}
+		else if(event->event_id == MPV_EVENT_NONE)
 		{
 			done = TRUE;
 		}
 
-		if(event)
+		if(event && !mpv->mpv_ctx)
 		{
-			if(mpv->event_callback)
-			{
-				mpv->event_callback
-					(event, mpv->event_callback_data);
-			}
-
-			if(mpv->mpv_ctx)
-			{
-				g_signal_emit_by_name
-					(mpv, "mpv-event", event->event_id);
-			}
-			else
-			{
-				done = TRUE;
-			}
+			done = TRUE;
 		}
 	}
 
 	return FALSE;
 }
 
-static void update_playlist(GmpvMpv *mpv)
-{
-	/* The length of "playlist//filename" including null-terminator (19)
-	 * plus the number of digits in the maximum value of 64 bit int (19).
-	 */
-	const gsize filename_prop_str_size = 38;
-	GtkListStore *store = gmpv_playlist_get_store(mpv->playlist);
-	gchar *filename_prop_str = g_malloc(filename_prop_str_size);
-	gboolean iter_end = FALSE;
-	GtkTreeIter iter;
-	mpv_node mpv_playlist;
-	gint playlist_count;
-	gint i;
-
-	mpv_check_error(mpv_get_property(	mpv->mpv_ctx,
-						"playlist",
-						MPV_FORMAT_NODE,
-						&mpv_playlist ));
-	playlist_count = mpv_playlist.u.list->num;
-
-	gtk_tree_model_get_iter_first(GTK_TREE_MODEL(store), &iter);
-
-	for(i = 0; i < playlist_count; i++)
-	{
-		mpv_node_list *prop_list = mpv_playlist.u.list->values[i].u.list;
-		gchar *uri = NULL;
-		gchar *title = NULL;
-		gchar *name = NULL;
-
-		for(gint j = 0; j < prop_list->num; j++)
-		{
-			const gchar *key = prop_list->keys[j];
-			const mpv_node value = prop_list->values[j];
-
-			if(g_strcmp0(key, "filename") == 0)
-			{
-				g_assert(value.format == MPV_FORMAT_STRING);
-				uri = value.u.string;
-			}
-			else if(g_strcmp0(key, "title") == 0)
-			{
-				g_assert(value.format == MPV_FORMAT_STRING);
-				title = value.u.string;
-			}
-		}
-
-		name = title?g_strdup(title):get_name_from_path(uri);
-
-		/* Overwrite current entry if it doesn't match the new value */
-		if(!iter_end)
-		{
-			gchar *old_name = NULL;
-			gchar *old_uri = NULL;
-			gboolean name_update;
-			gboolean uri_update;
-
-			gtk_tree_model_get
-				(	GTK_TREE_MODEL(store), &iter,
-					PLAYLIST_NAME_COLUMN, &old_name,
-					PLAYLIST_URI_COLUMN, &old_uri, -1 );
-
-			name_update = (g_strcmp0(name, old_name) != 0);
-			uri_update = (g_strcmp0(uri, old_uri) != 0);
-
-			/* Only set the name if either the title can be
-			 * retrieved or the name is unset. This preserves the
-			 * correct title if it becomes unavailable later such as
-			 * when restarting mpv.
-			 */
-			if(name_update && (!old_name || title || uri_update))
-			{
-				gtk_list_store_set
-					(	store, &iter,
-						PLAYLIST_NAME_COLUMN, name, -1 );
-			}
-
-			if(uri_update)
-			{
-				gtk_list_store_set
-					(	store, &iter,
-						PLAYLIST_URI_COLUMN, uri, -1 );
-			}
-
-			iter_end = !gtk_tree_model_iter_next
-					(GTK_TREE_MODEL(store), &iter);
-
-			g_free(old_name);
-			g_free(old_uri);
-		}
-		/* Append entries to the playlist if there are fewer entries in
-		 * the playlist widget than mpv's playlist.
-		 */
-		else
-		{
-			gmpv_playlist_append(mpv->playlist, name, uri);
-		}
-
-		g_free(name);
-	}
-
-	/* If there are more entries in the playlist widget than mpv's playlist,
-	 * remove the excess entries from the playlist widget.
-	 */
-	if(!iter_end)
-	{
-		while(gtk_list_store_remove(store, &iter));
-	}
-
-	g_free(filename_prop_str);
-	mpv_free_node_contents(&mpv_playlist);
-}
-
 static gint apply_args(mpv_handle *mpv_ctx, gchar *args)
 {
 	gchar *opt_begin = args?strstr(args, "--"):NULL;
@@ -607,6 +680,102 @@ static void load_input_conf(GmpvMpv *mpv, const gchar *input_conf)
 	fclose(tmp_file);
 }
 
+static void add_file_to_playlist(GmpvMpv *mpv, const gchar *uri)
+{
+	GmpvPlaylistEntry *entry = gmpv_playlist_entry_new(uri, NULL);
+
+	g_ptr_array_add(mpv->playlist, entry);
+	g_signal_emit_by_name(mpv, "mpv-prop-change", "playlist", NULL);
+}
+
+static void update_playlist(GmpvMpv *mpv)
+{
+	const mpv_node_list *org_list;
+	mpv_node playlist;
+
+	g_ptr_array_set_size(mpv->playlist, 0);
+	gmpv_mpv_get_property(mpv, "playlist", MPV_FORMAT_NODE, &playlist);
+
+	org_list = playlist.u.list;
+
+	if(playlist.format == MPV_FORMAT_NODE_ARRAY)
+	{
+		for(gint i = 0; i < org_list->num; i++)
+		{
+			GmpvPlaylistEntry *entry;
+
+			entry = parse_playlist_entry(org_list->values[i].u.list);
+			g_ptr_array_add(mpv->playlist, entry);
+		}
+
+		mpv_free_node_contents(&playlist);
+	}
+}
+
+static void update_metadata(GmpvMpv *mpv)
+{
+	mpv_node_list *org_list = NULL;
+	mpv_node metadata;
+
+	g_ptr_array_set_size(mpv->metadata, 0);
+	gmpv_mpv_get_property(mpv, "metadata", MPV_FORMAT_NODE, &metadata);
+	org_list = metadata.u.list;
+
+	if(metadata.format == MPV_FORMAT_NODE_MAP && org_list->num > 0)
+	{
+		for(gint i = 0; i < org_list->num; i++)
+		{
+			const gchar *key;
+			mpv_node value;
+
+			key = org_list->keys[i];
+			value = org_list->values[i];
+
+			if(value.format == MPV_FORMAT_STRING)
+			{
+				GmpvMetadataEntry *entry;
+
+				entry =	gmpv_metadata_entry_new
+					(key, value.u.string);
+
+				g_ptr_array_add(mpv->metadata, entry);
+			}
+			else
+			{
+				g_warning(	"Ignored metadata field %s "
+						"with unexpected format %d",
+						key,
+						value.format );
+			}
+		}
+
+		mpv_free_node_contents(&metadata);
+	}
+}
+
+static void update_track_list(GmpvMpv *mpv)
+{
+	mpv_node_list *org_list = NULL;
+	mpv_node track_list;
+
+	g_ptr_array_set_size(mpv->track_list, 0);
+	gmpv_mpv_get_property(mpv, "track-list", MPV_FORMAT_NODE, &track_list);
+	org_list = track_list.u.list;
+
+	if(track_list.format == MPV_FORMAT_NODE_ARRAY)
+	{
+		for(gint i = 0; i < org_list->num; i++)
+		{
+			GmpvTrack *entry =	parse_track_entry
+						(org_list->values[i].u.list);
+
+			g_ptr_array_add(mpv->track_list, entry);
+		}
+
+		mpv_free_node_contents(&track_list);
+	}
+}
+
 static void gmpv_mpv_class_init(GmpvMpvClass* klass)
 {
 	GObjectClass *obj_class = G_OBJECT_CLASS(klass);
@@ -625,13 +794,6 @@ static void gmpv_mpv_class_init(GmpvMpvClass* klass)
 			G_PARAM_READWRITE );
 	g_object_class_install_property(obj_class, PROP_WID, pspec);
 
-	pspec = g_param_spec_pointer
-		(	"playlist",
-			"Playlist",
-			"GmpvPlaylist to use for storage",
-			G_PARAM_READWRITE );
-	g_object_class_install_property(obj_class, PROP_PLAYLIST, pspec);
-
 	g_signal_new(	"mpv-init",
 			G_TYPE_FROM_CLASS(klass),
 			G_SIGNAL_RUN_FIRST,
@@ -641,7 +803,7 @@ static void gmpv_mpv_class_init(GmpvMpvClass* klass)
 			g_cclosure_marshal_VOID__VOID,
 			G_TYPE_NONE,
 			0 );
-	g_signal_new(	"mpv-error",
+	g_signal_new(	"error",
 			G_TYPE_FROM_CLASS(klass),
 			G_SIGNAL_RUN_FIRST,
 			0,
@@ -660,17 +822,37 @@ static void gmpv_mpv_class_init(GmpvMpvClass* klass)
 			g_cclosure_marshal_VOID__VOID,
 			G_TYPE_NONE,
 			0 );
-	g_signal_new(	"mpv-event",
+	g_signal_new(	"mpv-prop-change",
 			G_TYPE_FROM_CLASS(klass),
 			G_SIGNAL_RUN_FIRST,
 			0,
 			NULL,
 			NULL,
-			g_cclosure_marshal_VOID__ENUM,
+			g_cclosure_gen_marshal_VOID__STRING_POINTER,
 			G_TYPE_NONE,
-			1,
-			G_TYPE_INT );
-	g_signal_new(	"mpv-prop-change",
+			2,
+			G_TYPE_STRING,
+			G_TYPE_POINTER );
+
+	g_signal_new(	"idle",
+			G_TYPE_FROM_CLASS(klass),
+			G_SIGNAL_RUN_FIRST,
+			0,
+			NULL,
+			NULL,
+			g_cclosure_marshal_VOID__VOID,
+			G_TYPE_NONE,
+			0 );
+	g_signal_new(	"load",
+			G_TYPE_FROM_CLASS(klass),
+			G_SIGNAL_RUN_FIRST,
+			0,
+			NULL,
+			NULL,
+			g_cclosure_marshal_VOID__VOID,
+			G_TYPE_NONE,
+			0 );
+	g_signal_new(	"message",
 			G_TYPE_FROM_CLASS(klass),
 			G_SIGNAL_RUN_FIRST,
 			0,
@@ -680,13 +862,40 @@ static void gmpv_mpv_class_init(GmpvMpvClass* klass)
 			G_TYPE_NONE,
 			1,
 			G_TYPE_STRING );
+	g_signal_new(	"autofit",
+			G_TYPE_FROM_CLASS(klass),
+			G_SIGNAL_RUN_FIRST,
+			0,
+			NULL,
+			NULL,
+			g_cclosure_marshal_VOID__DOUBLE,
+			G_TYPE_NONE,
+			1,
+			G_TYPE_DOUBLE );
+	g_signal_new(	"shutdown",
+			G_TYPE_FROM_CLASS(klass),
+			G_SIGNAL_RUN_FIRST,
+			0,
+			NULL,
+			NULL,
+			g_cclosure_marshal_VOID__VOID,
+			G_TYPE_NONE,
+			0 );
 }
 
 static void gmpv_mpv_init(GmpvMpv *mpv)
 {
 	mpv->mpv_ctx = mpv_create();
 	mpv->opengl_ctx = NULL;
-	mpv->playlist = NULL;
+	mpv->playlist = g_ptr_array_new_full(	1,
+						(GDestroyNotify)
+						gmpv_playlist_entry_free );
+	mpv->metadata = g_ptr_array_new_full(	1,
+						(GDestroyNotify)
+						gmpv_metadata_entry_free );
+	mpv->track_list = g_ptr_array_new_full(	1,
+						(GDestroyNotify)
+						gmpv_track_free );
 	mpv->tmp_input_file = NULL;
 	mpv->log_level_list = NULL;
 	mpv->autofit_ratio = -1;
@@ -698,38 +907,18 @@ static void gmpv_mpv_init(GmpvMpv *mpv)
 	mpv->state.new_file = TRUE;
 	mpv->state.init_load = TRUE;
 
+	mpv->init_vo_config = TRUE;
 	mpv->force_opengl = FALSE;
+	mpv->use_opengl = FALSE;
 	mpv->glarea = NULL;
 	mpv->wid = -1;
-	mpv->event_callback_data = NULL;
 	mpv->opengl_cb_callback_data = NULL;
-	mpv->event_callback = NULL;
 	mpv->opengl_cb_callback = NULL;
 }
 
-GmpvMpv *gmpv_mpv_new(GmpvPlaylist *playlist, gint64 wid)
+GmpvMpv *gmpv_mpv_new(gint64 wid)
 {
-	return GMPV_MPV_OBJ(g_object_new(	gmpv_mpv_get_type(),
-						"playlist", playlist,
-						"wid", wid,
-						NULL ));
-}
-
-void mpv_check_error(int status)
-{
-	void *array[10];
-	size_t size;
-
-	if(status < 0)
-	{
-		size = (size_t)backtrace(array, 10);
-
-		g_critical("MPV API error: %s\n", mpv_error_string(status));
-
-		backtrace_symbols_fd(array, (int)size, STDERR_FILENO);
-
-		exit(EXIT_FAILURE);
-	}
+	return GMPV_MPV_OBJ(g_object_new(gmpv_mpv_get_type(), "wid", wid, NULL));
 }
 
 inline const GmpvMpvState *gmpv_mpv_get_state(GmpvMpv *mpv)
@@ -742,16 +931,6 @@ inline GmpvGeometry *gmpv_mpv_get_geometry(GmpvMpv *mpv)
 	return mpv->geometry;
 }
 
-inline gdouble gmpv_mpv_get_autofit_ratio(GmpvMpv *mpv)
-{
-	return mpv->autofit_ratio;
-}
-
-inline GmpvPlaylist *gmpv_mpv_get_playlist(GmpvMpv *mpv)
-{
-	return mpv->playlist;
-}
-
 inline mpv_handle *gmpv_mpv_get_mpv_handle(GmpvMpv *mpv)
 {
 	return mpv->mpv_ctx;
@@ -762,6 +941,26 @@ inline mpv_opengl_cb_context *gmpv_mpv_get_opengl_cb_context(GmpvMpv *mpv)
 	return mpv->opengl_ctx;
 }
 
+inline gboolean gmpv_mpv_get_use_opengl_cb(GmpvMpv *mpv)
+{
+	return mpv->use_opengl;
+}
+
+GPtrArray *gmpv_mpv_get_metadata(GmpvMpv *mpv)
+{
+	return mpv->metadata;
+}
+
+GPtrArray *gmpv_mpv_get_playlist(GmpvMpv *mpv)
+{
+	return mpv->playlist;
+}
+
+GPtrArray *gmpv_mpv_get_track_list(GmpvMpv *mpv)
+{
+	return mpv->track_list;
+}
+
 void gmpv_mpv_initialize(GmpvMpv *mpv)
 {
 	GSettings *main_settings = g_settings_new(CONFIG_ROOT);
@@ -775,16 +974,18 @@ void gmpv_mpv_initialize(GmpvMpv *mpv)
 		const gchar *name;
 		const gchar *value;
 	}
-	options[] = {	{"osd-level", "1"},
+	options[] = {	{"vo", "opengl,vdpau,vaapi,xv,x11,opengl-cb,"},
+			{"osd-level", "1"},
 			{"softvol", "yes"},
-			{"force-window", "yes"},
+			{"force-window", "immediate"},
 			{"input-default-bindings", "yes"},
 			{"audio-client-name", ICON_NAME},
 			{"title", "${media-title}"},
 			{"autofit-larger", "75%"},
 			{"window-scale", "1"},
-			{"pause", "no"},
+			{"pause", "yes"},
 			{"ytdl", "yes"},
+			{"load-scripts", "no"},
 			{"osd-bar", "no"},
 			{"input-cursor", "no"},
 			{"cursor-autohide", "no"},
@@ -845,7 +1046,7 @@ void gmpv_mpv_initialize(GmpvMpv *mpv)
 		const gchar *msg
 			= _("Failed to apply one or more MPV options.");
 
-		g_signal_emit_by_name(mpv, "mpv-error", msg);
+		g_signal_emit_by_name(mpv, "error", msg);
 	}
 
 	if(mpv->force_opengl || mpv->wid <= 0)
@@ -862,26 +1063,35 @@ void gmpv_mpv_initialize(GmpvMpv *mpv)
 		mpv_set_option(mpv->mpv_ctx, "wid", MPV_FORMAT_INT64, &mpv->wid);
 	}
 
-	mpv_observe_property(mpv->mpv_ctx, 0, "aid", MPV_FORMAT_INT64);
+	mpv_observe_property(mpv->mpv_ctx, 0, "aid", MPV_FORMAT_STRING);
+	mpv_observe_property(mpv->mpv_ctx, 0, "vid", MPV_FORMAT_STRING);
+	mpv_observe_property(mpv->mpv_ctx, 0, "sid", MPV_FORMAT_STRING);
 	mpv_observe_property(mpv->mpv_ctx, 0, "chapters", MPV_FORMAT_INT64);
 	mpv_observe_property(mpv->mpv_ctx, 0, "core-idle", MPV_FORMAT_FLAG);
+	mpv_observe_property(mpv->mpv_ctx, 0, "idle-active", MPV_FORMAT_FLAG);
 	mpv_observe_property(mpv->mpv_ctx, 0, "fullscreen", MPV_FORMAT_FLAG);
 	mpv_observe_property(mpv->mpv_ctx, 0, "pause", MPV_FORMAT_FLAG);
+	mpv_observe_property(mpv->mpv_ctx, 0, "loop", MPV_FORMAT_STRING);
 	mpv_observe_property(mpv->mpv_ctx, 0, "duration", MPV_FORMAT_DOUBLE);
 	mpv_observe_property(mpv->mpv_ctx, 0, "media-title", MPV_FORMAT_STRING);
+	mpv_observe_property(mpv->mpv_ctx, 0, "metadata", MPV_FORMAT_NODE);
+	mpv_observe_property(mpv->mpv_ctx, 0, "playlist", MPV_FORMAT_NODE);
+	mpv_observe_property(mpv->mpv_ctx, 0, "playlist-count", MPV_FORMAT_INT64);
 	mpv_observe_property(mpv->mpv_ctx, 0, "playlist-pos", MPV_FORMAT_INT64);
+	mpv_observe_property(mpv->mpv_ctx, 0, "speed", MPV_FORMAT_DOUBLE);
 	mpv_observe_property(mpv->mpv_ctx, 0, "track-list", MPV_FORMAT_NODE);
+	mpv_observe_property(mpv->mpv_ctx, 0, "vo-configured", MPV_FORMAT_FLAG);
 	mpv_observe_property(mpv->mpv_ctx, 0, "volume", MPV_FORMAT_DOUBLE);
 	mpv_set_wakeup_callback(mpv->mpv_ctx, wakeup_callback, mpv);
-	mpv_check_error(mpv_initialize(mpv->mpv_ctx));
-
+	mpv_initialize(mpv->mpv_ctx);
 
 	mpv_version = gmpv_mpv_get_property_string(mpv, "mpv-version");
 	current_vo = gmpv_mpv_get_property_string(mpv, "current-vo");
+	mpv->use_opengl = !current_vo;
 
 	g_info("Using %s", mpv_version);
 
-	if(current_vo && !GDK_IS_X11_DISPLAY(gdk_display_get_default()))
+	if(!mpv->use_opengl && !GDK_IS_X11_DISPLAY(gdk_display_get_default()))
 	{
 		g_info(	"The chosen vo is %s but the display is not X11; "
 			"forcing --vo=opengl-cb and resetting",
@@ -906,8 +1116,7 @@ void gmpv_mpv_initialize(GmpvMpv *mpv)
 					MPV_FORMAT_DOUBLE,
 					&volume );
 
-		/* The vo should be opengl-cb if current_vo is NULL*/
-		if(!current_vo)
+		if(mpv->use_opengl)
 		{
 			mpv->opengl_ctx =	mpv_get_sub_api
 						(	mpv->mpv_ctx,
@@ -976,16 +1185,12 @@ void gmpv_mpv_reset(GmpvMpv *mpv)
 	/* Reset mpv->mpv_ctx */
 	mpv->state.ready = FALSE;
 
-	mpv_check_error(gmpv_mpv_command(mpv, quit_cmd));
+	gmpv_mpv_command(mpv, quit_cmd);
 	gmpv_mpv_quit(mpv);
 
 	mpv->mpv_ctx = mpv_create();
 	gmpv_mpv_initialize(mpv);
 
-	gmpv_mpv_set_event_callback
-		(	mpv,
-			mpv->event_callback,
-			mpv->event_callback_data );
 	gmpv_mpv_set_opengl_cb_callback
 		(	mpv,
 			mpv->opengl_cb_callback,
@@ -997,17 +1202,9 @@ void gmpv_mpv_reset(GmpvMpv *mpv)
 	{
 		if(mpv->state.loaded)
 		{
-			gint rc;
-
-			rc =	mpv_request_event
-				(mpv->mpv_ctx, MPV_EVENT_FILE_LOADED, 0);
-			mpv_check_error(rc);
-
-			gmpv_mpv_load(mpv, NULL, FALSE, TRUE);
-
-			rc =	mpv_request_event
-				(mpv->mpv_ctx, MPV_EVENT_FILE_LOADED, 1);
-			mpv_check_error(rc);
+			mpv_request_event(mpv->mpv_ctx, MPV_EVENT_FILE_LOADED, 0);
+			load_from_playlist(mpv);
+			mpv_request_event(mpv->mpv_ctx, MPV_EVENT_FILE_LOADED, 1);
 		}
 
 		if(playlist_pos_rc >= 0 && playlist_pos > 0)
@@ -1048,106 +1245,97 @@ void gmpv_mpv_quit(GmpvMpv *mpv)
 	mpv->mpv_ctx = NULL;
 }
 
-void gmpv_mpv_load(	GmpvMpv *mpv,
-			const gchar *uri,
-			gboolean append,
-			gboolean update )
+void gmpv_mpv_load_track(GmpvMpv *mpv, const gchar *uri, TrackType type)
 {
-	const gchar *load_cmd[] = {"loadfile", NULL, NULL, NULL};
-	GtkListStore *playlist_store = gmpv_playlist_get_store(mpv->playlist);
-	GtkTreeIter iter;
-	gboolean empty;
+	const gchar *cmd[3] = {NULL};
+	gchar *path = g_filename_from_uri(uri, NULL, NULL);
 
-	g_info(	"Loading file (append=%s, update=%s): %s",
-		append?"TRUE":"FALSE",
-		update?"TRUE":"FALSE",
-		uri?:"<PLAYLIST_ITEMS>" );
+	if(type == TRACK_TYPE_AUDIO)
+	{
+		cmd[0] = "audio-add";
+	}
+	else if(type == TRACK_TYPE_SUBTITLE)
+	{
+		cmd[0] = "sub-add";
+	}
+	else
+	{
+		g_assert_not_reached();
+	}
 
-	empty = !gtk_tree_model_get_iter_first
-			(GTK_TREE_MODEL(playlist_store), &iter);
+	cmd[1] = path?:uri;
 
-	load_cmd[2] = (append && !empty)?"append":"replace";
+	g_debug("Loading external track %s with type %d", cmd[1], type);
+	gmpv_mpv_command(mpv, cmd);
 
-	if(!append && uri && update)
+	g_free(path);
+}
+
+void gmpv_mpv_load_file(GmpvMpv *mpv, const gchar *uri, gboolean append)
+{
+	gchar *path = get_path_from_uri(uri);
+	const gchar *load_cmd[] = {"loadfile", path, NULL, NULL};
+	gint64 playlist_count = 0;
+
+	g_assert(uri);
+	g_info(	"Loading file (append=%s): %s", append?"TRUE":"FALSE", uri);
+
+	mpv_get_property(	mpv->mpv_ctx,
+				"playlist-count",
+				MPV_FORMAT_INT64,
+				&playlist_count );
+
+	load_cmd[2] = (append && playlist_count > 0)?"append":"replace";
+
+	if(!append)
 	{
 		mpv->state.new_file = TRUE;
 		mpv->state.loaded = FALSE;
-	}
-
-	if(!uri)
-	{
-		gboolean append = FALSE;
-		gboolean rc;
 
 		if(!mpv->state.init_load)
 		{
 			gmpv_mpv_set_property_flag(mpv, "pause", FALSE);
 		}
-
-		rc = gtk_tree_model_get_iter_first
-			(GTK_TREE_MODEL(playlist_store), &iter);
-
-		while(rc)
-		{
-			gchar *uri;
-
-			gtk_tree_model_get(	GTK_TREE_MODEL(playlist_store),
-						&iter,
-						PLAYLIST_URI_COLUMN,
-						&uri,
-						-1 );
-
-			/* append = FALSE only on first iteration */
-			gmpv_mpv_load(mpv, uri, append, FALSE);
-
-			append = TRUE;
-
-			rc = gtk_tree_model_iter_next
-				(GTK_TREE_MODEL(playlist_store), &iter);
-
-			g_free(uri);
-		}
 	}
 
-	if(uri && playlist_store)
+	g_assert(mpv->mpv_ctx);
+	mpv_request_event(mpv->mpv_ctx, MPV_EVENT_END_FILE, 0);
+	mpv_command(mpv->mpv_ctx, load_cmd);
+	mpv_request_event(mpv->mpv_ctx, MPV_EVENT_END_FILE, 1);
+
+	g_free(path);
+}
+
+void gmpv_mpv_load(GmpvMpv *mpv, const gchar *uri, gboolean append)
+{
+	const gchar *subtitle_exts[] = SUBTITLE_EXTS;
+
+	if(extension_matches(uri, subtitle_exts))
 	{
-		gchar *path = get_path_from_uri(uri);
+		gmpv_mpv_load_track(mpv, uri, TRACK_TYPE_SUBTITLE);
+	}
+	else
+	{
+		gboolean idle_active = FALSE;
 
-		load_cmd[1] = path;
+		mpv_get_property(	mpv->mpv_ctx,
+					"idle-active",
+					MPV_FORMAT_FLAG,
+					&idle_active );
 
-		if(!append)
+		if(idle_active)
 		{
-			mpv->state.loaded = FALSE;
-
-			if(!mpv->state.init_load)
+			if(!append)
 			{
-				gmpv_mpv_set_property_flag
-					(mpv, "pause", FALSE);
+				g_ptr_array_set_size(mpv->playlist, 0);
 			}
-		}
 
-		if(update)
+			add_file_to_playlist(mpv, uri);
+		}
+		else
 		{
-			gchar *name = get_name_from_path(path);
-
-			gmpv_playlist_append(mpv->playlist, name, uri);
-
-			g_free(name);
+			gmpv_mpv_load_file(mpv, uri, append);
 		}
-
-		g_assert(mpv->mpv_ctx);
-
-		mpv_check_error(mpv_request_event(	mpv->mpv_ctx,
-							MPV_EVENT_END_FILE,
-							0 ));
-
-		mpv_check_error(mpv_command(mpv->mpv_ctx, load_cmd));
-
-		mpv_check_error(mpv_request_event(	mpv->mpv_ctx,
-							MPV_EVENT_END_FILE,
-							1 ));
-
-		g_free(path);
 	}
 }
 
@@ -1155,62 +1343,3 @@ void gmpv_mpv_free(gpointer data)
 {
 	mpv_free(data);
 }
-
-void gmpv_mpv_load_list(	GmpvMpv *mpv,
-				const gchar **uri_list,
-				gboolean append,
-				gboolean update )
-{
-	static const char *const sub_exts[] = SUBTITLE_EXTS;
-
-	for(gint i = 0; uri_list[i]; i++)
-	{
-		const gchar *ext = strrchr(uri_list[i], '.');
-		gboolean subtitle = FALSE;
-
-		/* Only start checking the extension if there is at
-		 * least one character after the dot.
-		 */
-		if(ext && ++ext)
-		{
-			const gchar *const *cur = sub_exts;
-
-			/* Check if the file extension matches one of the
-			 * supported subtitle formats.
-			 */
-			while(*cur && g_strcmp0(ext, *(cur++)) != 0);
-
-			subtitle = !!(*cur);
-		}
-
-		/* Only attempt to load file as subtitle if there
-		 * already is a file loaded. Try to load the file as a
-		 * media file otherwise.
-		 */
-		if(ext && subtitle && gmpv_mpv_get_state(mpv)->loaded)
-		{
-			const gchar *cmd[] = {"sub-add", NULL, NULL};
-			gchar *path;
-
-			/* Convert to path if possible to get rid of
-			 * percent encoding.
-			 */
-			path = g_filename_from_uri(uri_list[i], NULL, NULL);
-			cmd[1] = path?:uri_list[i];
-
-			g_debug("Loading external subtitle: %s", cmd[1]);
-			gmpv_mpv_command(mpv, cmd);
-
-			g_free(path);
-		}
-		else
-		{
-			gboolean empty = gmpv_playlist_empty(mpv->playlist);
-
-			gmpv_mpv_load(	mpv,
-					uri_list[i],
-					((append && !empty) || i != 0),
-					TRUE );
-		}
-	}
-}
