@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2016 gnome-mpv
+ * Copyright (c) 2016-2017 gnome-mpv
  *
  * This file is part of GNOME MPV.
  *
@@ -20,6 +20,8 @@
 #ifndef MPV_OBJ_PRIVATE_H
 #define MPV_OBJ_PRIVATE_H
 
+#include <gtk/gtk.h>
+
 #include "gmpv_mpv_opt.h"
 
 G_BEGIN_DECLS
@@ -28,7 +30,6 @@ enum
 {
 	PROP_0,
 	PROP_WID,
-	PROP_PLAYLIST,
 	N_PROPERTIES
 };
 
@@ -38,17 +39,19 @@ struct _GmpvMpv
 	GmpvMpvState state;
 	mpv_handle *mpv_ctx;
 	mpv_opengl_cb_context *opengl_ctx;
-	GmpvPlaylist *playlist;
+	GPtrArray *playlist;
+	GPtrArray *metadata;
+	GPtrArray *track_list;
 	GtkGLArea *glarea;
 	gchar *tmp_input_file;
 	GSList *log_level_list;
 	gdouble autofit_ratio;
 	GmpvGeometry *geometry;
+	gboolean init_vo_config;
 	gboolean force_opengl;
+	gboolean use_opengl;
 	gint64 wid;
-	void *event_callback_data;
 	void *opengl_cb_callback_data;
-	void (*event_callback)(mpv_event *, void *data);
 	void (*opengl_cb_callback)(void *data);
 };
 
