@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2016 gnome-mpv
+ * Copyright (c) 2016-2017 gnome-mpv
  *
  * This file is part of GNOME MPV.
  *
@@ -30,6 +30,13 @@
 #include <gdk/gdkx.h>
 #endif
 
+enum
+{
+	PROP_0,
+	PROP_TITLE,
+	N_PROPERTIES
+};
+
 struct _GmpvVideoArea
 {
 	GtkOverlay parent_instance;
@@ -50,6 +57,14 @@ struct _GmpvVideoAreaClass
 	GtkOverlayClass parent_class;
 };
 
+static void set_property(	GObject *object,
+				guint property_id,
+				const GValue *value,
+				GParamSpec *pspec );
+static void get_property(	GObject *object,
+				guint property_id,
+				GValue *value,
+				GParamSpec *pspec );
 static void set_cursor_visible(GmpvVideoArea *area, gboolean visible);
 static gboolean timeout_handler(gpointer data);
 static gboolean motion_notify_handler(GtkWidget *widget, GdkEventMotion *event);
@@ -59,6 +74,48 @@ static gboolean fs_control_crossing_handler(	GtkWidget *widget,
 
 G_DEFINE_TYPE(GmpvVideoArea, gmpv_video_area, GTK_TYPE_OVERLAY)
 
+static void set_property(	GObject *object,
+				guint property_id,
+				const GValue *value,
+				GParamSpec *pspec )
+{
+	GmpvVideoArea *self = GMPV_VIDEO_AREA(object);
+
+	switch(property_id)
+	{
+		case PROP_TITLE:
+		g_object_set_property(	G_OBJECT(self->header_bar),
+					pspec->name,
+					value );
+		break;
+
+		default:
+		G_OBJECT_WARN_INVALID_PROPERTY_ID(object, property_id, pspec);
+		break;
+	}
+}
+
+static void get_property(	GObject *object,
+				guint property_id,
+				GValue *value,
+				GParamSpec *pspec )
+{
+	GmpvVideoArea *self = GMPV_VIDEO_AREA(object);
+
+	switch(property_id)
+	{
+		case PROP_TITLE:
+		g_object_get_property(	G_OBJECT(self->header_bar),
+					pspec->name,
+					value );
+		break;
+
+		default:
+		G_OBJECT_WARN_INVALID_PROPERTY_ID(object, property_id, pspec);
+		break;
+	}
+}
+
 static void set_cursor_visible(GmpvVideoArea *area, gboolean visible)
 {
 	GdkWindow *window;
@@ -87,21 +144,29 @@ static gboolean timeout_handler(gpointer data)
 	GmpvControlBox *control_box = GMPV_CONTROL_BOX(area->control_box);
 	GmpvHeaderBar *header_bar = GMPV_HEADER_BAR(area->header_bar);
 
-	if(area->fullscreen
+	if(control_box
 	&& !area->fs_control_hover
 	&& !gmpv_control_box_get_volume_popup_visible(control_box)
 	&& !gmpv_header_bar_get_open_button_popup_visible(header_bar)
 	&& !gmpv_header_bar_get_menu_button_popup_visible(header_bar))
 	{
-		set_cursor_visible(area, FALSE);
 		gtk_revealer_set_reveal_child
 			(GTK_REVEALER(area->control_box_revealer), FALSE);
 		gtk_revealer_set_reveal_child
 			(GTK_REVEALER(area->header_bar_revealer), FALSE);
 
+		set_cursor_visible(area, !area->fullscreen);
+		area->timeout_tag = 0;
+	}
+	else if(!control_box)
+	{
 		area->timeout_tag = 0;
 	}
 
+	/* Try again later if timeout_tag has not been cleared. This means that
+	 * either one of the popups is visible or the cursor is hovering over
+	 * the control box, preventing it from being hidden.
+	 */
 	return (area->timeout_tag != 0);
 }
 
@@ -113,12 +178,12 @@ static gboolean motion_notify_handler(GtkWidget *widget, GdkEventMotion *event)
 	cursor = gdk_cursor_new_from_name(gdk_display_get_default(), "default");
 	gdk_window_set_cursor(gtk_widget_get_window(widget), cursor);
 
-	if(area->fullscreen)
+	if(area->control_box)
 	{
 		gtk_revealer_set_reveal_child
 			(GTK_REVEALER(area->control_box_revealer), TRUE);
 		gtk_revealer_set_reveal_child
-			(GTK_REVEALER(area->header_bar_revealer), TRUE);
+			(GTK_REVEALER(area->header_bar_revealer), area->fullscreen);
 	}
 
 	if(area->timeout_tag > 0)
@@ -134,6 +199,15 @@ static gboolean motion_notify_handler(GtkWidget *widget, GdkEventMotion *event)
 		->motion_notify_event(widget, event);
 }
 
+static gboolean render_handler(	GtkGLArea *gl_area,
+				GdkGLContext *context,
+				gpointer data )
+{
+	g_signal_emit_by_name(data, "render");
+
+	return FALSE;
+}
+
 static void notify_handler(	GObject *gobject,
 				GParamSpec *pspec,
 				gpointer data )
@@ -168,9 +242,31 @@ static gboolean fs_control_crossing_handler(	GtkWidget *widget,
 
 static void gmpv_video_area_class_init(GmpvVideoAreaClass *klass)
 {
+	GObjectClass *obj_class = G_OBJECT_CLASS(klass);
 	GtkWidgetClass *wgt_class = GTK_WIDGET_CLASS(klass);
+	GParamSpec *pspec = NULL;
 
+	obj_class->set_property = set_property;
+	obj_class->get_property = get_property;
 	wgt_class->motion_notify_event = motion_notify_handler;
+
+	g_signal_new(	"render",
+			G_TYPE_FROM_CLASS(klass),
+			G_SIGNAL_RUN_FIRST,
+			0,
+			NULL,
+			NULL,
+			g_cclosure_marshal_VOID__VOID,
+			G_TYPE_NONE,
+			0 );
+
+	pspec = g_param_spec_string
+		(	"title",
+			"Title",
+			"The title of the header bar",
+			NULL,
+			G_PARAM_READWRITE );
+	g_object_class_install_property(obj_class, PROP_TITLE, pspec);
 }
 
 static void gmpv_video_area_init(GmpvVideoArea *area)
@@ -224,6 +320,10 @@ static void gmpv_video_area_init(GmpvVideoArea *area)
 	gtk_widget_hide(area->header_bar_revealer);
 	gtk_widget_set_no_show_all(area->header_bar_revealer, TRUE);
 
+	g_signal_connect(	area->gl_area,
+				"render",
+				G_CALLBACK(render_handler),
+				area );
 	g_signal_connect(	area->header_bar_revealer,
 				"notify::reveal-child",
 				G_CALLBACK(notify_handler),
@@ -259,15 +359,10 @@ GtkWidget *gmpv_video_area_new()
 }
 
 void gmpv_video_area_update_track_list(	GmpvVideoArea *area,
-					const GSList *audio_list,
-					const GSList *video_list,
-					const GSList *sub_list )
+					const GPtrArray *track_list )
 {
 	gmpv_header_bar_update_track_list
-		(	GMPV_HEADER_BAR(area->header_bar),
-			audio_list,
-			video_list,
-			sub_list );
+		(GMPV_HEADER_BAR(area->header_bar), track_list);
 }
 
 void gmpv_video_area_set_fullscreen_state(	GmpvVideoArea *area,
@@ -277,7 +372,6 @@ void gmpv_video_area_set_fullscreen_state(	GmpvVideoArea *area,
 	{
 		area->fullscreen = fullscreen;
 
-		gtk_widget_set_visible(area->control_box_revealer, fullscreen);
 		gtk_widget_hide(area->header_bar_revealer);
 		set_cursor_visible(area, !fullscreen);
 
@@ -322,6 +416,11 @@ void gmpv_video_area_set_use_opengl(GmpvVideoArea *area, gboolean use_opengl)
 			use_opengl?area->gl_area:area->draw_area );
 }
 
+void gmpv_video_area_queue_render(GmpvVideoArea *area)
+{
+	gtk_gl_area_queue_render(GTK_GL_AREA(area->gl_area));
+}
+
 GtkDrawingArea *gmpv_video_area_get_draw_area(GmpvVideoArea *area)
 {
 	return GTK_DRAWING_AREA(area->draw_area);
